/*
Firmware for model railway control interface

ATMega pinout:
D0 - serial RX
D1 - serial TX
D2 - Polarity control
D3 - Power control

Connections for Pololu high-powered motor driver:
D2 <-> DIR
D3 <-> PWM

Connections for S88 bus:
D4 <-> RESET
D5 <-> LOAD
D6 <-> CLOCK
D7 <-> DATA

ADC connections:
ADC0 - current sensor (adjusted for 185 mV/A, centered at Vcc/2)
ADC1 - input voltage (adjusted for divisor of 11)

This is not a complete control device.  Rather, it is a low-level bus driver,
intended to be controlled with a computer program.  In particular, no state is
kept about locomotives or solenoids.  It is up to the controlling program to
maintain a refresh cycle, turn off solenoids after a sufficient delay, etc.

Commands are received as a packetized stream over the serial interface.  Each
packet begins with a header byte, which is bitwise not of the packet's length,
not including the header byte itself.  The longest allowed packet is 15 bytes.
Thus, the high four bits of the header byte are always ones.  The first data
byte specifies the command, and the rest of the packet is command arguments.

After a command has been processed, a similarly packetized response is sent.

Track output is asynchronous up to a single packet.  If an output command is
received while a previous packet is still being sent to the track, execution
blocks until the operation completes.  The command returns as soon as sending
its data to the track can start.
*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include "clock.h"
#include "interface.h"
#include "monitor.h"
#include "output.h"
#include "s88.h"

int main(void)
{
	clock_start();
	output_init();
	interface_init();
	monitor_init();
	s88_init();

	sei();

	while(1)
	{
		interface_check();
		monitor_check();
		s88_check();
	}

	return 0;
}
